#include <iostream>
#include <fstream>
#include <string>
#include <ctime>
#include <map>
#include <cassert>
#include <boost/regex.hpp>
#include "colors.h"
using namespace std;

const char* re = "[0-9,A-F]{6}";
boost::regex expression(re);

string StringToUpper(string s)
{
    transform(s.begin(), s.end(), s.begin(), std::toupper);
    return s;
}

void Help(void)
{
    cout << "Change the colours in Notepadd++ stylers.xml file" << endl << endl
         << "Usage: xNpp stylers.xml" << endl;
}

string ChangeColour(string s)
{
    static map <string, string> colours;
    map <string, string>::const_iterator x;
    typedef pair <string, string> ColPair;

    assert(s.length() == 6);
    s = StringToUpper(s);
    if (colours.empty())
    {
        // black to white
        assert(colours.find(black) == colours.end());
        colours.insert(ColPair(black, white));
        
        // white to black.
        assert(colours.find(white) == colours.end());
        colours.insert(ColPair(white, black));
        
        // silver to gray30.
        assert(colours.find(silver) == colours.end());
        colours.insert(ColPair(silver, gray35));
        
        // red to red.
        assert(colours.find(red) == colours.end());
        colours.insert(ColPair(red, red));
        
        // maroon to navajowhite.
        assert(colours.find("800000") == colours.end());
        colours.insert(ColPair("800000", navajowhite));
        
        // near aliceblue to midnightblue.
        assert(colours.find("E8E8FF") == colours.end());
        colours.insert(ColPair("E8E8FF", midnightblue));

        // near slateblue to cornflowerblue. 
        assert(colours.find("8000FF") == colours.end());
        colours.insert(ColPair("8000FF", cornflowerblue));
        
        // near cyan to darkcyan.
        assert(colours.find("80FFFF") == colours.end());
        colours.insert(ColPair("80FFFF", darkcyan));

        // gray to gray70.
        assert(colours.find(gray) == colours.end());
        colours.insert(ColPair(gray, gray70));

        // near whitesmoke to black.
        assert(colours.find("F3F3F3") == colours.end());
        colours.insert(ColPair("F3F3F3", black));

        // near lightcoral to indianred.
        assert(colours.find("FF8080") == colours.end());
        colours.insert(ColPair("FF8080", indianred));

        // near firebrick to tomato.
        assert(colours.find("C00000") == colours.end());
        colours.insert(ColPair("C00000", tomato));

        // near saddlebrown to burlywood.
        assert(colours.find("804000") == colours.end());
        colours.insert(ColPair("804000", burlywood));

        // blue to deepskyblue.
        assert(colours.find(blue) == colours.end());
        colours.insert(ColPair(blue, deepskyblue));

        // near darkorange to orange.
        assert(colours.find("FF8000") == colours.end());
        colours.insert(ColPair("FF8000", orange));

        // navy to lightblue.
        assert(colours.find(navy) == colours.end());
        colours.insert(ColPair(navy, lightblue));

        // green to mediumseagreen.
        assert(colours.find(green) == colours.end());
        colours.insert(ColPair(green, mediumseagreen));

        // near ? to mediumturquoise (or lightseagreen).
        assert(colours.find("008080") == colours.end());
        colours.insert(ColPair("008080", mediumturquoise));

        // near darkorchid to orchid.
        assert(colours.find("A001D6") == colours.end());
        colours.insert(ColPair("A001D6", orchid));

        // near darkmagenta to magenta.
        assert(colours.find("9500A4") == colours.end());
        colours.insert(ColPair("9500A4", magenta));

        // near lightblue to darkblue.
        assert(colours.find("A6CAF0") == colours.end());
        colours.insert(ColPair("A6CAF0", darkblue));

        // near lightyellow to maroon.
        assert(colours.find("FEFDE0") == colours.end());
        colours.insert(ColPair("FEFDE0", maroon));

        // near lightyellow to saddlebrown.
        assert(colours.find("FDF8E3") == colours.end());
        colours.insert(ColPair("FDF8E3", saddlebrown));

        // near floralwhite to gray15.
        assert(colours.find("FEFCF5") == colours.end());
        colours.insert(ColPair("FEFCF5", gray15));

        // near aliceblue to midnightblue.
        assert(colours.find("F2F4FF") == colours.end());
        colours.insert(ColPair("F2F4FF", midnightblue));

        // near powderblue to slateblue.
        assert(colours.find("C4F9FD") == colours.end());
        colours.insert(ColPair("C4F9FD", slateblue));

        // yellow to yellow.
        assert(colours.find(yellow) == colours.end());
        colours.insert(ColPair(yellow, darkgreen));

        // near olive to olive.
        assert(colours.find("808040") == colours.end());
        colours.insert(ColPair("808040", olive));

        // near whitesmoke to gray30.
        assert(colours.find("E9E9E9") == colours.end());
        colours.insert(ColPair("E9E9E9", gray30));

        // near gainsboro to gray5.
        assert(colours.find("E4E4E4") == colours.end());
        colours.insert(ColPair("E4E4E4", gray5));

        // near deeppink to deeppink.
        assert(colours.find("FF0080") == colours.end());
        colours.insert(ColPair("FF0080", deeppink));

        // near hotpink to lightpink.
        assert(colours.find("FF80C0") == colours.end());
        colours.insert(ColPair("FF80C0", lightpink));

        // near dodgerblue to dodgerblue.
        assert(colours.find("0080FF") == colours.end());
        colours.insert(ColPair("0080FF", dodgerblue));

        // olive to greenyellow.
        assert(colours.find(olive) == colours.end());
        colours.insert(ColPair(olive, greenyellow));

        // near ? to lightsteelblue.
        assert(colours.find("8080C0") == colours.end());
        colours.insert(ColPair("8080C0", lightsteelblue));

        // near honeydew to gray5.
        assert(colours.find("ECFFEA") == colours.end());
        colours.insert(ColPair("ECFFEA", gray5));

        // near lemonchiffon to navy.
        assert(colours.find("FFFFCC") == colours.end());
        colours.insert(ColPair("FFFFCC", midnightblue));

        // near mediumslateblue to cornflowerblue.
        assert(colours.find("8080FF") == colours.end());
        colours.insert(ColPair("8080FF", cornflowerblue));

        // near teal to paleturquoise.
        assert(colours.find("408080") == colours.end());
        colours.insert(ColPair("408080", paleturquoise));

        // near greenyellow to greenyellow.
        assert(colours.find("B5E71F") == colours.end());
        colours.insert(ColPair("B5E71F", greenyellow));

        // near dodgerblue to cornflowerblue.
        assert(colours.find("0080C0") == colours.end());
        colours.insert(ColPair("0080C0", cornflowerblue));

        // magenta to magenta.
        assert(colours.find(magenta) == colours.end());
        colours.insert(ColPair(magenta, magenta));

        // near papayawhip to gray5.
        assert(colours.find("FFF0D8") == colours.end());
        colours.insert(ColPair("FFF0D8", gray5));

        //dark maroon to orange.
        assert(colours.find("600000") == colours.end());
        colours.insert(ColPair("600000", orange));

        // near rosybrown to saddlebrown.
        assert(colours.find("A08080") == colours.end());
        colours.insert(ColPair("A08080", saddlebrown));

        // light palegreen to darkgreen.
        assert(colours.find("C0FFC0") == colours.end());
        colours.insert(ColPair("C0FFC0", darkgreen));

        // dark darkgreen to palegreen.
        assert(colours.find("004000") == colours.end());
        colours.insert(ColPair("004000", palegreen));

        // near magenta to magenta.
        assert(colours.find("FF80FF") == colours.end());
        colours.insert(ColPair("FF80FF", magenta));

        // near lightcyan to darkcyan.
        assert(colours.find("E1FFF3") == colours.end());
        colours.insert(ColPair("E1FFF3", darkcyan));

        // dark saddlebrown to sandybrown.
        assert(colours.find("804040") == colours.end());
        colours.insert(ColPair("804040", sandybrown));

        // cyan to darkcyan.
        assert(colours.find("00FFFF") == colours.end());
        colours.insert(ColPair("00FFFF", darkcyan));

        // near lightyellow to black.
        assert(colours.find("FFFFD9") == colours.end());
        colours.insert(ColPair("FFFFD9", black));

        // near coral to coral.
        assert(colours.find("FF8040") == colours.end());
        colours.insert(ColPair("FF8040", coral));

        // near darkred to sandybrown.
        assert(colours.find("95004A") == colours.end());
        colours.insert(ColPair("95004A", sandybrown));

        // ? to gray5.
        assert(colours.find("EFEFEF") == colours.end());
        colours.insert(ColPair("EFEFEF", gray5));

        // ? to black.
        assert(colours.find("FDFFEC") == colours.end());
        colours.insert(ColPair("FDFFEC", black));

        // near yellow to darkgreen.
        assert(colours.find("FFFF80") == colours.end());
        colours.insert(ColPair("FFFF80", darkgreen));

        // ? to gray5.
        assert(colours.find("EEEEEE") == colours.end());
        colours.insert(ColPair("EEEEEE",gray5));

        // near darkblue to skyblue.
        assert(colours.find("0000A0") == colours.end());
        colours.insert(ColPair("0000A0", skyblue));

        // ? to gray5.
        assert(colours.find("FCFFF0") == colours.end());
        colours.insert(ColPair("FCFFF0", "03000F"));

        // ? to gray5.
        assert(colours.find("FFEEEC") == colours.end());
        colours.insert(ColPair("FFEEEC", gray5));

        // near orchid to orchid.
        assert(colours.find("CF34CF") == colours.end());
        colours.insert(ColPair("CF34CF", orchid));

        // ? to darkgoldenrod.
        assert(colours.find("FFC000") == colours.end());
        colours.insert(ColPair("FFC000", darkgoldenrod));

        // lime to lime.
        assert(colours.find(lime) == colours.end());
        colours.insert(ColPair(lime, lime));

        // ? to gray5.
        assert(colours.find("F8FEDE") == colours.end());
        colours.insert(ColPair("F8FEDE", gray5));

        // Added for npp v. 4.2.2

        // purple to magenta
        assert(colours.find(purple) == colours.end());
        colours.insert(ColPair(purple, magenta));

        // near chocolate to chocolate
        assert(colours.find("CA6500") == colours.end());
        colours.insert(ColPair("CA6500", chocolate));

        // ? to gray10
        assert(colours.find("D5FFD5") == colours.end());
        colours.insert(ColPair("D5FFD5", chocolate));

        // near sandybrown to saddlebrown
        assert(colours.find("FFB56A") == colours.end());
        colours.insert(ColPair("FFB56A", saddlebrown));
    }

    x = colours.find(s);
    if (x == colours.end())
    {
        cerr << endl 
             << "Unable to find the colour " << s << endl;
        return "??????";
    }
    else
    {
        assert(x->second.length() == 6);
        assert(StringToUpper(x->second) == x->second);
        return x->second;
    }
}

int main(int argc, char* argv[])
{
    if (argc != 2)
    {
        Help();
        return 1;
    }
    ifstream in(argv[1]);
    if(!in)
    {
        cerr << "Could not open file "<< argv[1] << endl;
        return 1;
    }
    string line;
    if (getline(in, line))
    {
        time_t ltime;
        time(&ltime);
        string now(ctime(&ltime));
        string::size_type i = now.find_last_of('\n');
        if (i != string::npos)
            now = now.substr(0, i);
        cout << line << endl 
             << "<!-- Colours changed with xNpp on " << now << " -->" << endl;;
    }
    while(getline(in, line))
    {
        string::const_iterator start, end;
        start = line.begin();
        end = line.end();
        boost::match_results<string::const_iterator> what;
        boost::match_flag_type flags = boost::match_default;
        if (boost::regex_search(start, end, what, expression, flags))
        {
            string what_2;
            do
            {
                string what_1 = what[-1];   // Everything from the start
                                            // of the input string to the
                                            // start of this match.
                cout << what_1;
                assert(what.size() == 1);
                string what0 = what[0]; // what[0] contains the whole string.
                cout << ChangeColour(what0);

                what_2 = what[-2];  // Everything from the end of the match,
                                    // to the end of the input string.
                start = what[0].second;
                // Update flags:
                flags |= boost::match_prev_avail | boost::match_not_bob;
            } while (boost::regex_search(start, end, what, expression, flags));
            cout << what_2 << endl;
        }
        else
            cout << line << endl;
   }
   return 0;
}
